package File::Winny::Cache;
# http://www.binzume.net/library/winny02.html

use strict;
use Crypt::RC4;
use Digest::MD5;

use constant CRYPT_KEY1 => 'header';
use constant CRYPT_KEY2 => 'adsfu6';

### <<<CONSTRUCTOR>>>
sub new {
    my ($class, $param) = @_;

    $param ||= {};
    bless $param, $class;
}

### Get cache information
sub info {
    my ($self) = @_;

    if (defined (my $filename = $self->{file})) {
        my $buffer;
        if (open (my $fh, "<$filename")) {
            binmode $fh;
            sysread $fh, $buffer, 0x140;

            my $version = substr $buffer, 0x10, 4;
            $version = RC4( CRYPT_KEY1, $version );
            $version = _endian ($version);
            $version =~ s/(.)/sprintf '%02x',ord($1)/egs;
            $self->{version} = hex ($version);

            my $buffer2 = substr $buffer, 0x14, 0x12C;
            $buffer2 = RC4( CRYPT_KEY2, $buffer2 );

            my $size = substr $buffer2, 0x14 - 0x14, 4;
            $size = _endian ($size);
            $size =~ s/(.)/sprintf '%02x',ord($1)/egs;
            $self->{size} = hex ($size);
            $self->{block_num} = int (($self->{size} + 65535) / 65536);

            my $hash = substr $buffer2, 0x20 - 0x14, 16;
            $hash =~ s/(.)/sprintf '%02x',ord($1)/egs;
            $self->{hash} = $hash;

            my $namelen = substr $buffer2, 0x3e - 0x14, 1;
            my $namesum = substr $buffer2, 0x40 - 0x14, 1;
            my $name = substr $buffer2, 0x42 - 0x14, ord($namelen);
            $self->{name} = RC4( $namesum, $name );

            seek $fh, 0, 0;
            sysread $fh, $buffer, 0x140 + $self->{block_num};
            $buffer2 = substr $buffer, 0x14, 0x12C + $self->{block_num};
            $buffer2 = RC4( CRYPT_KEY2, $buffer2 );
            my @block = ();
            foreach (0..$self->{block_num}-1) {
                $block[$_] = ord (substr $buffer2, 0x12C + $_, 1);
            }
            $self->{block} = \@block;

            close $fh;
        };
    }
    $self;
}

### is complete cache ?
sub is_complete {
    my ($self) = @_;

    $self->info if !defined $self->{block};
    map { return 0 if !$_; } @{$self->{block}};
    1;
}

### Convert the cache and retrieve the crypted content
sub convert_to {
    my ($self, $filename) = @_;

    $self->info if !defined $self->{block};

    if (open (my $src, "<". $self->{file})) {
        binmode $src;
        if (open (my $dst, ">$filename")) {
            binmode $dst;
            foreach (0..$self->{block_num}-1) {
                seek $src, 0x140 + $self->{block_num} + (65536 + 16) * $_, 0;

                my $md5_hash;
                sysread $src, $md5_hash, 16;
                # $md5_hash =~ s/(.)/sprintf '%02x',ord($1)/egs; print "$md5_hash\n";

                my $buffer;
                my $block_size = ($_ == $self->{block_num}-1)
                    ? $self->{size} - 65536 * $_
                    : 65536;
                sysread $src, $buffer, $block_size;
                if ($_ == 0) {
                    $buffer = RC4( CRYPT_KEY1, $buffer );
                }
                else {
                    my $crypt_key = sprintf '%d %s', $_, $self->{hash};
                    $buffer = RC4( $crypt_key, $buffer );
                }
                # my $md5_hash2 = Digest::MD5::md5($buffer); $md5_hash2 =~ s/(.)/sprintf '%02x',ord($1)/egs; print " - $md5_hash2\n";

                print $dst $buffer;
            }
            close $dst;
        }
        close $src;
    }
}

### Switch the endian
sub _endian { join '', reverse split //, $_[0]; }

1;